"""PyAuthz Django integration — first-class authorization for Django.

Provides middleware, decorators, and a permission backend for integrating
PyAuthz authorization into Django applications.

Quick Example:
    # settings.py
    MIDDLEWARE = ["pyauthz_django.PyAuthzMiddleware", ...]
    PYAUTHZ_POLICY_DIR = "policies/"

    # views.py
    from pyauthz_django import requires

    @requires("documents:delete")
    def delete_document(request, doc_id):
        ...
"""

from pyauthz_django.decorator import requires
from pyauthz_django.middleware import PyAuthzMiddleware

__all__ = [
    "PyAuthzMiddleware",
    "requires",
]
